# mypy: allow-untyped-decorators
# mypy: allow-untyped-defs
from typing import cast, List, Optional, Tuple, Union

import torch
from torch import Tensor

from torch.optim.optimizer import (
    _capturable_doc,
    _default_to_fused_or_foreach,
    _device_dtype_check_for_fused,
    _differentiable_doc,
    _disable_dynamo_if_unsupported,
    _foreach_doc,
    _fused_doc,
    _get_capturable_supported_devices,
    _get_scalar_dtype,
    _get_value,
    _maximize_doc,
    _stack_if_compiling,
    _use_grad_for_differentiable,
    _view_as_real,
    DeviceDict,
    Optimizer,
    ParamsT,
)


__all__ = ["ADOPT", "adopt"]


class ADOPT(Optimizer):
    ### "adapted" from https://github.com/iShohei220/adopt/blob/main/adopt.py
    def __init__(
        self,
        params: ParamsT,
        lr: Union[float, Tensor] = 1e-3,
        betas: Tuple[float, float] = (0.9, 0.9999),
        eps: float = 1e-6,
        weight_decay: float = 0.0,
        decoupled: bool = False,
        *,
        foreach: Optional[bool] = None,
        maximize: bool = False,
        capturable: bool = False,
        differentiable: bool = False,
        fused: Optional[bool] = None,
    ):
        if isinstance(lr, Tensor):
            if foreach and not capturable:
                raise ValueError(
                    "lr as a Tensor is not supported for `capturable=False` and `foreach=True`"
                )
            if lr.numel() != 1:
                raise ValueError("Tensor lr must be 1-element")
        if not 0.0 <= lr:
            raise ValueError(f"Invalid learning rate: {lr}")
        if not 0.0 <= eps:
            raise ValueError(f"Invalid epsilon value: {eps}")
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError(f"Invalid beta parameter at index 0: {betas[0]}")
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError(f"Invalid beta parameter at index 1: {betas[1]}")
        if not 0.0 <= weight_decay:
            raise ValueError(f"Invalid weight_decay value: {weight_decay}")

        defaults = dict(
            lr=lr,
            betas=betas,
            eps=eps,
            weight_decay=weight_decay,
            decoupled=decoupled,
            maximize=maximize,
            foreach=foreach,
            capturable=capturable,
            differentiable=differentiable,
            fused=fused,
        )
        super().__init__(params, defaults)

        if fused:
            # TODO: support fused
            raise RuntimeError("`fused` is not currently supported")

            if differentiable:
                raise RuntimeError("`fused` does not support `differentiable`")
            self._step_supports_amp_scaling = True
            # TODO(crcrpar): [low prec params & their higher prec copy]
            # Support AMP with FP16/BF16 model params which would need
            # higher prec copy of params to do update math in higher prec to
            # alleviate the loss of information.
            if foreach:
                raise RuntimeError("`fused` and `foreach` cannot be `True` together.")

    def __setstate__(self, state):
        super().__setstate__(state)
        for group in self.param_groups:
            group.setdefault("maximize", False)
            group.setdefault("foreach", None)
            group.setdefault("capturable", False)
            group.setdefault("differentiable", False)
            fused = group.setdefault("fused", None)
            for p in group["params"]:
                p_state = self.state.get(p, [])
                if len(p_state) != 0 and not torch.is_tensor(p_state["step"]):
                    step_val = float(p_state["step"])
                    p_state["step"] = (
                        torch.tensor(
                            step_val,
                            dtype=_get_scalar_dtype(is_fused=fused),
                            device=p.device,
                        )
                        if group["capturable"] or group["fused"]
                        else torch.tensor(step_val, dtype=_get_scalar_dtype())
                    )

    def _init_group(
        self, group, params_with_grad, grads, exp_avgs, exp_avg_sqs, state_steps
    ):
        has_complex = False
        for p in group["params"]:
            if p.grad is not None:
                has_complex |= torch.is_complex(p)
                params_with_grad.append(p)
                if p.grad.is_sparse:
                    raise RuntimeError("ADOPT does not support sparse gradients")
                grads.append(p.grad)

                state = self.state[p]
                # Lazy state initialization
                if len(state) == 0:
                    if group["fused"]:
                        _device_dtype_check_for_fused(p)
                    # note(crcrpar): [special device hosting for step]
                    # Deliberately host `step` on CPU if both capturable and fused are off.
                    # This is because kernel launches are costly on CUDA and XLA.
                    state["step"] = (
                        torch.zeros(
                            (),
                            dtype=_get_scalar_dtype(is_fused=group["fused"]),
                            device=p.device,
                        )
                        if group["capturable"] or group["fused"]
                        else torch.tensor(0.0, dtype=_get_scalar_dtype())
                    )
                    # Exponential moving average of gradient values
                    state["exp_avg"] = torch.zeros_like(
                        p, memory_format=torch.preserve_format
                    )
                    # Exponential moving average of squared gradient values
                    state["exp_avg_sq"] = torch.zeros_like(
                        p, memory_format=torch.preserve_format
                    )

                exp_avgs.append(state["exp_avg"])
                exp_avg_sqs.append(state["exp_avg_sq"])

                if group["differentiable"] and state["step"].requires_grad:
                    raise RuntimeError(
                        "`requires_grad` is not supported for `step` in differentiable mode"
                    )

                # Foreach without capturable does not support a tensor lr
                if (
                    group["foreach"]
                    and torch.is_tensor(group["lr"])
                    and not group["capturable"]
                ):
                    raise RuntimeError(
                        "lr as a Tensor is not supported for capturable=False and foreach=True"
                    )

                state_steps.append(state["step"])
        return has_complex

    @_use_grad_for_differentiable
    def step(self, closure=None):
        """Perform a single optimization step.

        Args:
            closure (Callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        self._cuda_graph_capture_health_check()

        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            params_with_grad: List[Tensor] = []
            grads: List[Tensor] = []
            exp_avgs: List[Tensor] = []
            exp_avg_sqs: List[Tensor] = []
            state_steps: List[Tensor] = []
            beta1, beta2 = group["betas"]

            has_complex = self._init_group(
                group, params_with_grad, grads, exp_avgs, exp_avg_sqs, state_steps
            )

            adopt(
                params_with_grad,
                grads,
                exp_avgs,
                exp_avg_sqs,
                state_steps,
                has_complex=has_complex,
                beta1=beta1,
                beta2=beta2,
                lr=group["lr"],
                weight_decay=group["weight_decay"],
                decoupled=group["decoupled"],
                eps=group["eps"],
                maximize=group["maximize"],
                foreach=group["foreach"],
                capturable=group["capturable"],
                differentiable=group["differentiable"],
                fused=group["fused"],
                grad_scale=getattr(self, "grad_scale", None),
                found_inf=getattr(self, "found_inf", None),
            )

        return loss


def _single_tensor_adopt(
    params: List[Tensor],
    grads: List[Tensor],
    exp_avgs: List[Tensor],
    exp_avg_sqs: List[Tensor],
    state_steps: List[Tensor],
    grad_scale: Optional[Tensor],
    found_inf: Optional[Tensor],
    *,
    has_complex: bool,
    beta1: float,
    beta2: float,
    lr: Union[float, Tensor],
    weight_decay: float,
    decoupled: bool,
    eps: float,
    maximize: bool,
    capturable: bool,
    differentiable: bool,
):
    assert grad_scale is None and found_inf is None

    if torch.jit.is_scripting():
        # this assert is due to JIT being dumb and not realizing that the ops below
        # have overloads to handle both float and Tensor lrs, so we just assert it's
        # a float since most people using JIT are using floats
        assert isinstance(lr, float)

    for i, param in enumerate(params):
        grad = grads[i] if not maximize else -grads[i]
        exp_avg = exp_avgs[i]
        exp_avg_sq = exp_avg_sqs[i]
        step_t = state_steps[i]

        # If compiling, the compiler will handle cudagraph checks, see note [torch.compile x capturable]
        if not torch._utils.is_compiling() and capturable:
            capturable_supported_devices = _get_capturable_supported_devices()
            assert (
                param.device.type == step_t.device.type
                and param.device.type in capturable_supported_devices
            ), f"If capturable=True, params and state_steps must be on supported devices: {capturable_supported_devices}."

        # update step
        step_t += 1

        if weight_decay != 0:
            if decoupled:
                param.add_(param, alpha=-lr * weight_decay)
            else:
                grad = grad.add(param, alpha=weight_decay)

        if torch.is_complex(param):
            grad = torch.view_as_real(grad)
            if exp_avg is not None:
                exp_avg = torch.view_as_real(exp_avg)
            if exp_avg_sq is not None:
                exp_avg_sq = torch.view_as_real(exp_avg_sq)
            param = torch.view_as_real(param)

        step = step_t if capturable or differentiable else _get_value(step_t)
        if step == 1:
            exp_avg_sq.addcmul_(grad, grad.conj())
            continue

        denom = torch.clamp(exp_avg_sq.sqrt(), eps)
        if step == 2:
            exp_avg.addcdiv_(grad, denom)
        else:
            exp_avg.mul_(beta1).addcdiv_(grad, denom, value=1 - beta1)

        param.add_(exp_avg, alpha=-lr)
        exp_avg_sq.mul_(beta2).addcmul_(grad, grad.conj(), value=1 - beta2)


def _multi_tensor_adopt(
    params: List[Tensor],
    grads: List[Tensor],
    exp_avgs: List[Tensor],
    exp_avg_sqs: List[Tensor],
    state_steps: List[Tensor],
    grad_scale: Optional[Tensor],
    found_inf: Optional[Tensor],
    *,
    has_complex: bool,
    beta1: float,
    beta2: float,
    lr: Union[float, Tensor],
    weight_decay: float,
    decoupled: bool,
    eps: float,
    maximize: bool,
    capturable: bool,
    differentiable: bool,
):
    if len(params) == 0:
        return

    if isinstance(lr, Tensor) and not capturable:
        raise RuntimeError(
            "lr as a Tensor is not supported for capturable=False and foreach=True"
        )

    # If compiling, the compiler will handle cudagraph checks, see note [torch.compile x capturable]
    if not torch._utils.is_compiling() and capturable:
        capturable_supported_devices = _get_capturable_supported_devices(
            supports_xla=False
        )
        assert all(
            p.device.type == step.device.type
            and p.device.type in capturable_supported_devices
            for p, step in zip(params, state_steps)
        ), f"If capturable=True, params and state_steps must be on supported devices: {capturable_supported_devices}."

    assert grad_scale is None and found_inf is None

    assert not differentiable, "_foreach ops don't support autograd"

    grouped_tensors = Optimizer._group_tensors_by_device_and_dtype(
        [params, grads, exp_avgs, exp_avg_sqs, state_steps]  # type: ignore[list-item]
    )
    for (
        device_params_,
        device_grads_,
        device_exp_avgs_,
        device_exp_avg_sqs_,
        device_state_steps_,
    ), _ in grouped_tensors.values():
        device_params = cast(List[Tensor], device_params_)
        device_grads = cast(List[Tensor], device_grads_)
        device_exp_avgs = cast(List[Tensor], device_exp_avgs_)
        device_exp_avg_sqs = cast(List[Tensor], device_exp_avg_sqs_)
        device_state_steps = cast(List[Tensor], device_state_steps_)

        # Handle complex parameters
        if has_complex:
            _view_as_real(
                device_params, device_grads, device_exp_avgs, device_exp_avg_sqs
            )

        if maximize:
            device_grads = torch._foreach_neg(device_grads)  # type: ignore[assignment]

        # Update steps
        # If steps are on CPU, foreach will fall back to the slow path, which is a for-loop calling t.add(1) over
        # and over. 1 will then be wrapped into a Tensor over and over again, which is slower than if we just
        # wrapped it once now. The alpha is required to assure we go to the right overload.
        if not torch._utils.is_compiling() and device_state_steps[0].is_cpu:
            torch._foreach_add_(
                device_state_steps, torch.tensor(1.0, device="cpu"), alpha=1.0
            )
        else:
            torch._foreach_add_(device_state_steps, 1)

        if weight_decay != 0:
            if decoupled:
                torch._foreach_add_(
                    device_params, device_params, alpha=-lr * weight_decay
                )
            else:
                # Re-use the intermediate memory (device_grads) already allocated for maximize
                if maximize:
                    torch._foreach_add_(device_grads, device_params, alpha=weight_decay)
                else:
                    device_grads = torch._foreach_add(  # type: ignore[assignment]
                        device_grads, device_params, alpha=weight_decay
                    )

        if device_state_steps[0] == 1:
            torch._foreach_addcmul_(device_exp_avg_sqs, device_grads, device_grads)
            continue

        exp_avg_sq_sqrt = torch._foreach_sqrt(device_exp_avg_sqs)
        exp_avg_sq_sqrt = torch._foreach_maximum(exp_avg_sq_sqrt, eps)

        if device_state_steps[0] == 2:
            torch._foreach_addcdiv_(device_exp_avgs, device_grads, exp_avg_sq_sqrt)
        else:
            torch._foreach_mul_(device_exp_avgs, beta1)
            torch._foreach_addcdiv_(
                device_exp_avgs, device_grads, exp_avg_sq_sqrt, value=1 - beta1
            )

        torch._foreach_add_(device_params, device_exp_avgs, alpha=-lr)
        torch._foreach_mul_(device_exp_avg_sqs, beta2)
        torch._foreach_addcmul_(
            device_exp_avg_sqs, device_grads, device_grads, value=1 - beta2
        )


@_disable_dynamo_if_unsupported(single_tensor_fn=_single_tensor_adopt)
def adopt(
    params: List[Tensor],
    grads: List[Tensor],
    exp_avgs: List[Tensor],
    exp_avg_sqs: List[Tensor],
    state_steps: List[Tensor],
    # kwonly args with defaults are not supported by functions compiled with torchscript issue #70627
    # setting this as kwarg for now as functional API is compiled by torch/distributed/optim
    foreach: Optional[bool] = None,
    capturable: bool = False,
    differentiable: bool = False,
    fused: Optional[bool] = None,
    grad_scale: Optional[Tensor] = None,
    found_inf: Optional[Tensor] = None,
    has_complex: bool = False,
    *,
    beta1: float,
    beta2: float,
    lr: Union[float, Tensor],
    weight_decay: float,
    decoupled: bool,
    eps: float,
    maximize: bool,
):
    r"""Functional API that performs ADOPT algorithm computation."""
    # Respect when the user inputs False/True for foreach or fused. We only want to change
    # the default when neither have been user-specified. Note that we default to foreach
    # and pass False to use_fused. This is not a mistake--we want to give the fused impl
    # bake-in time before making it the default, even if it is typically faster.
    if fused is None and foreach is None:
        _, foreach = _default_to_fused_or_foreach(
            params, differentiable, use_fused=False
        )
        # Do not flip on foreach for the unsupported case where lr is a Tensor and capturable=False.
        if foreach and isinstance(lr, Tensor) and not capturable:
            foreach = False
    if fused is None:
        fused = False
    if foreach is None:
        foreach = False

    # this check is slow during compilation, so we skip it
    # if it's strictly needed we can add this check back in dynamo
    if not torch._utils.is_compiling() and not all(
        isinstance(t, torch.Tensor) for t in state_steps
    ):
        raise RuntimeError(
            "API has changed, `state_steps` argument must contain a list of singleton tensors"
        )

    if foreach and torch.jit.is_scripting():
        raise RuntimeError("torch.jit.script not supported with foreach optimizers")
    if fused and torch.jit.is_scripting():
        raise RuntimeError("torch.jit.script not supported with fused optimizers")

    if fused and not torch.jit.is_scripting():
        func = _fused_adopt
    elif foreach and not torch.jit.is_scripting():
        func = _multi_tensor_adopt
    else:
        func = _single_tensor_adopt

    func(
        params,
        grads,
        exp_avgs,
        exp_avg_sqs,
        state_steps,
        has_complex=has_complex,
        beta1=beta1,
        beta2=beta2,
        lr=lr,
        weight_decay=weight_decay,
        decoupled=decoupled,
        eps=eps,
        maximize=maximize,
        capturable=capturable,
        differentiable=differentiable,
        grad_scale=grad_scale,
        found_inf=found_inf,
    )


def adopt_wrapper(parameters: dict) -> torch.optim.Optimizer:
    """
    Creates an AdEMAMix optimizer from the PyTorch `torch.optim` module using the input parameters.

    Args:
        parameters (dict): A dictionary containing the input parameters for the optimizer.

    Returns:
        torch.optim.Optimizer: An AdEMAMix optimizer.
    """

    return ADOPT(
        params=parameters["model_parameters"],
        lr=parameters.get("learning_rate", 1e-3),
        betas=parameters.get("betas", (0.9, 0.9999)),
        eps=parameters.get("eps", 1e-6),
        weight_decay=parameters.get("weight_decay", 0.0),
        decoupled=parameters["optimizer"].get("decoupled", False),
        foreach=parameters.get("foreach", None),
        maximize=parameters.get("maximize", False),
        capturable=parameters.get("capturable", False),
        differentiable=parameters.get("differentiable", False),
        fused=parameters.get("fused", None),
    )



from pydantic import BaseModel, ConfigDict
from GANDLF.configuration.user_defined_config import UserDefinedParameters


class ParametersConfiguration(BaseModel):
    model_config = ConfigDict(extra="allow")


class Parameters(ParametersConfiguration, UserDefinedParameters):
    pass



import lightning.pytorch as pl
from GANDLF.compute.generic import (
    TrainingSubsetDataParser,
    ValidationSubsetDataParser,
    TestSubsetDataParser,
    InferenceSubsetDataParserRadiology,
)
from torch.utils.data import DataLoader as TorchDataLoader
from copy import deepcopy


class GandlfTrainingDatamodule(pl.LightningDataModule):
    def __init__(self, data_dict_files: dict, parameters_dict: dict):
        super().__init__()

        # Batch size here and reinitialization of dataloader parsers is used
        # in automatic batch size tuning

        self.batch_size = parameters_dict["batch_size"]

        # This init procedure is extreme hack, but the only way to get around the
        # need to modify the parameters dict during the parsing procedure

        params = deepcopy(parameters_dict)

        train_subset_parser = TrainingSubsetDataParser(
            data_dict_files["training"], params
        )
        self.training_dataset = train_subset_parser.create_subset_dataset()
        params = train_subset_parser.get_params_extended_with_subset_data()

        val_subset_parser = ValidationSubsetDataParser(
            data_dict_files["validation"], params
        )
        self.validation_dataset = val_subset_parser.create_subset_dataset()
        params = val_subset_parser.get_params_extended_with_subset_data()

        testing_data = data_dict_files.get("testing", None)
        self.test_dataset = None
        if testing_data is not None:
            test_subset_parser = TestSubsetDataParser(testing_data, params)
            self.test_dataset = test_subset_parser.create_subset_dataset()
            params = test_subset_parser.get_params_extended_with_subset_data()

        self.parameters_dict = params

    def _get_dataloader(self, dataset, batch_size: int, shuffle: bool):
        return TorchDataLoader(
            dataset,
            batch_size=batch_size,
            shuffle=shuffle,
            num_workers=self.updated_parameters_dict.get("num_workers_dataloader", 1),
            pin_memory=self.updated_parameters_dict.get("pin_memory_dataloader", False),
            prefetch_factor=self.updated_parameters_dict.get(
                "prefetch_factor_dataloader", 2
            ),
        )

    @property
    def updated_parameters_dict(self):
        return self.parameters_dict

    def train_dataloader(self):
        self.updated_parameters_dict["batch_size"] = self.batch_size
        return self._get_dataloader(
            self.training_dataset, self.batch_size, shuffle=True
        )

    def val_dataloader(self):
        return self._get_dataloader(
            self.validation_dataset, batch_size=1, shuffle=False
        )

    def test_dataloader(self):
        if self.test_dataset is None:
            return None
        return self._get_dataloader(self.test_dataset, batch_size=1, shuffle=False)


class GandlfInferenceDatamodule(pl.LightningDataModule):
    def __init__(self, dataframe, parameters_dict):
        super().__init__()
        self.dataframe = dataframe
        params = deepcopy(parameters_dict)
        self.parameters_dict = params
        if self.parameters_dict["modality"] == "rad":
            inference_subset_data_parser_radiology = InferenceSubsetDataParserRadiology(
                self.dataframe, params
            )
            self.inference_dataset = (
                inference_subset_data_parser_radiology.create_subset_dataset()
            )

            self.parameters_dict = (
                inference_subset_data_parser_radiology.get_params_extended_with_subset_data()
            )

    @property
    def updated_parameters_dict(self):
        return self.parameters_dict

    def predict_dataloader(self):
        if self.parameters_dict["modality"] == "rad":
            return TorchDataLoader(
                self.inference_dataset,
                batch_size=1,
                shuffle=False,
                num_workers=self.updated_parameters_dict.get(
                    "num_workers_dataloader", 1
                ),
                pin_memory=self.updated_parameters_dict.get(
                    "pin_memory_dataloader", False
                ),
                prefetch_factor=self.updated_parameters_dict.get(
                    "prefetch_factor_dataloader", 2
                ),
            )
        elif self.parameters_dict["modality"] in ["path", "histo"]:
            return self.dataframe.iterrows()



from pydantic import BaseModel, ConfigDict, Field, AliasChoices, model_validator
from typing_extensions import Any, Literal, Self


class ThresholdConfig(BaseModel):
    min: int = Field()
    max: int = Field()


class ClipConfig(BaseModel):
    min: int = Field()
    max: int = Field()


class RescaleConfig(BaseModel):
    in_min_max: list[float] = Field(default=[15, 125])
    out_min_max: list[float] = Field(default=[0, 1])
    percentiles: list[float] = Field(default=[5, 95])


class HistogramMatchingConfig(BaseModel):
    num_hist_level: int = Field(default=1024)
    num_match_points: int = Field(default=16)
    target: Any = Field(default=None)


class ResampleMinConfig(BaseModel):
    resolution: list[float] = Field(default=None)


class ResampleConfig(BaseModel):
    resolution: list[float] = Field(default=None)


class StainNormalizationConfig(BaseModel):
    target: Any = Field()
    extractor: Literal["vahadane", "ruifrok", "macenko"] = Field(default="ruifrok")


class PreProcessingConfig(BaseModel):
    model_config = ConfigDict(extra="forbid", exclude_none=True)
    to_canonical: Any = Field(default=None)
    threshold: ThresholdConfig = Field(default=None)
    clip: ClipConfig = Field(default=None)
    clamp: ClipConfig = Field(default=None)
    crop_external_zero_planes: Any = Field(default=None)
    crop: list[int] = Field(default=None)
    centercrop: list[int] = Field(default=None)
    normalize_by_val: Any = Field(default=None)
    normalize_imagenet: Any = Field(default=None)
    normalize_standardize: Any = Field(default=None)
    normalize_div_by_255: Any = Field(default=None)
    normalize: Any = Field(default=None)
    normalize_nonZero: Any = Field(
        default=None,
        validation_alias=AliasChoices("normalize_nonZero", "normalize_nonzero"),
    )
    normalize_nonZero_masked: Any = Field(
        default=None,
        validation_alias=AliasChoices(
            "normalize_nonZero_masked", "normalize_nonzero_masked"
        ),
    )
    rescale: RescaleConfig = Field(default=None)
    rgba2rgb: Any = Field(
        default=None,
        validation_alias=AliasChoices("rgba2rgb", "rgbatorgb", "rgba_to_rgb"),
    )
    rgb2rgba: Any = Field(
        default=None,
        validation_alias=AliasChoices("rgb2rgba", "rgbtorgba", "rgb_to_rgba"),
    )
    histogram_matching: HistogramMatchingConfig = Field(default=None)
    histogram_equalization: HistogramMatchingConfig = Field(default=None)
    adaptive_histogram_equalization: Any = Field(default=None)
    resample: ResampleConfig = Field(default=None)
    resize_image: list[int] = Field(
        default=None,
        validation_alias=AliasChoices(
            "resize_image", "resize", "resize_image", "resize_images"
        ),
    )
    resize_patch: list[int] = Field(default=None)
    stain_normalization: StainNormalizationConfig = Field(default=None)
    resample_min: ResampleMinConfig = Field(
        default=None, validation_alias=AliasChoices("resample_min", "resample_minimum")
    )

    @model_validator(mode="after")
    def pre_processing_validate(self) -> Self:
        if self.adaptive_histogram_equalization is not None:
            self.histogram_matching = HistogramMatchingConfig(target="adaptive")
            self.adaptive_histogram_equalization = None
        return self
